from fastapi.staticfiles import StaticFiles
from starlette.types import Scope
from starlette.responses import Response


class SpaStaticFiles(StaticFiles):
    async def get_response(self, path: str, scope: Scope) -> Response:
        response = await super().get_response(path, scope)
        if response.status_code == 404:
            response = await super().get_response(path=".", scope=scope)
        return response



from typing import Any, Dict, List
from fastapi import APIRouter, Request, HTTPException
import pyjson5 as json5
from fact_explorer.app.api.constants import API_RESPONSES

from fact_explorer.app.business import search
from fact_explorer.app.business.crypto_helpers import cryptoshred_enabled
from fact_explorer.app.business.registry import schema_registry_enabled
from fact_explorer.app.entities.fact import FactOut
from fact_explorer.config import get_logger

log = get_logger("api.router")

router = APIRouter(
    prefix="/api",
    responses=API_RESPONSES,
)


@router.get("/search", response_model=List[FactOut])
async def root(
    request: Request,
    skip: int = 0,
    limit: int = 20,
    hq: str = "{}",
    pq: str = "{}",
    decrypt: bool = False,
) -> List[Dict[str, Any]]:
    """
    Query the Facts received. Results are ordered by when the where received.
    Newest first.

    - `skip`: skip the X newest entries
    - `limit`: display X facts
    - `hq`: the header query. Use query by example here
    Just send a dict of what you are looking for.
    - `pq`: the payload query. Use query by example here.
    Just send a dict of what you are looking for.
    - `decrypt`: whether or not to decrypt cryptoshred information
    (Caution might increase response time significantly)
    """
    try:
        header_query = json5.decode(hq)
        payload_query = json5.decode(pq)
    except json5.Json5DecoderException as e:
        log.debug(e)
        raise HTTPException(status_code=400, detail="Invalid JSON5 parameters")

    result = await search.strict(
        db=request.app.state.db,
        skip=skip,
        limit=limit,
        hq=header_query,
        pq=payload_query,
        decrypt=decrypt,
    )
    return result


@router.get("/last", response_model=List[FactOut])
async def last(
    request: Request, minutes: int = 15, decrypt: bool = False
) -> List[Dict[str, Any]]:
    """
    Displays the Facts received in the last X minutes. Results are ordered,
    newest first.

    Beware that this is a very expensive operations in terms of memory and time.
    Use with care, you will kill the application if you use to large values.

    - `minutes`: How many minutes from now backwards you want to get.
    - `decrypt`: whether or not to decrypt cryptoshred information
    (Caution might increase response time significantly)
    """
    result = await search.time_interval(
        db=request.app.state.db,
        minutes=minutes,
        decrypt=decrypt,
    )
    return result


@router.get("/features")
async def features() -> Dict[str, bool]:
    """
    Returns an object containing as keys the set of optional features.
    The values indicate if that feature is currently available.
    """
    return {
        "cryptoshred": await cryptoshred_enabled(),
        "schema_registry": await schema_registry_enabled(),
    }



from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional
from fact_explorer.app.business.crypto_helpers import decrypt_result
from fact_explorer.app.db.db_session import Database


async def strict(
    *,
    db: Database,
    skip: int = 0,
    limit: int = 20,
    hq: Optional[Dict[str, Any]] = None,
    pq: Optional[Dict[str, Any]] = None,
    decrypt: bool = False,
) -> List[Dict[str, Any]]:

    if hq is None:
        hq = {}
    if pq is None:
        pq = {}

    response = await db.fetch_by_example(
        offset=skip, limit=limit, header_example=hq, payload_example=pq
    )
    if decrypt:
        result = await decrypt_result(data=response)
    else:
        result = response
    return result


async def time_interval(
    *,
    db: Database,
    start: Optional[datetime] = None,
    minutes: int = 15,
    decrypt: bool = False,
) -> List[Dict[str, Any]]:
    if not start:
        start = datetime.now()
    minutes_ago = start - timedelta(minutes=minutes)
    response = await db.fetch_by_time(until=minutes_ago)

    if decrypt:
        result = await decrypt_result(data=response)
    else:
        result = response
    return result



from typing import Any, Dict, List, Optional, Union
from fastapi import APIRouter, Query, Request
from fact_explorer.app.api.constants import API_RESPONSES
from fact_explorer.config import get_logger
from fact_explorer.app.business import registry as business


log = get_logger("api.registry")

router = APIRouter(prefix="/api/registry", responses=API_RESPONSES)


@router.get("/namespaces")
async def namespaces(
    request: Request, with_types: bool = Query(False, alias="with-types")
) -> Union[List[str], Dict[str, List[str]]]:
    """
    Query available namespaces. Can be used to also get event types for all namespaces.

    - `with_types`: also return the types as part of the response. Depending on your
    setup you might be better of with one or two requests here.
    """
    return await business.namespaces(with_types=with_types)


@router.get("/schema")
async def schema(
    request: Request, namespace: str, type: str, version: Optional[int] = None
) -> Dict[str, Any]:
    """
    For a given namespace/type/version tuple, return the stored json schema.

    - `namespace`: The namespace to look in.
    - `type`: The type to look for.
    - `version`: The version of the schema. (Optional - latest by default)
    """
    return await business.schema(namespace=namespace, type=type, version=version)



from typing import List, Optional
from uuid import UUID
from pydantic import BaseModel, Extra


class HeaderMeta(BaseModel):
    class Config:
        extra = Extra.allow

    _ts: int
    _ser: int


class FactHeader(BaseModel):
    class Config:
        extra = Extra.allow

    id: UUID
    ns: str
    aggIds: Optional[List[UUID]] = None
    type: Optional[str] = None
    meta: Optional[HeaderMeta] = None


class FactPayload(BaseModel):
    class Config:
        extra = Extra.allow


class FactOut(BaseModel):
    header: FactHeader
    payload: FactPayload



from typing import Any, Dict, Optional, Union, List

from fact_explorer.config import get_configuration, get_logger

log = get_logger("business.registry")
configuration = get_configuration()

try:
    from schema_registry.registries.HttpRegistry import EnrichingHttpSchemaRegistry
    from schema_registry.registries.entities import VersionedType

    registry = EnrichingHttpSchemaRegistry(url=configuration.schema_registry_url)

    async def namespaces(
        with_types: bool = False,
    ) -> Union[List[str], Dict[str, VersionedType]]:
        if with_types:
            res = {}
            for namespace in registry.namespaces:

                res[namespace.name] = {
                    "title": namespace.title,
                    "types": list(registry.types(namespace=namespace.name)),
                }
            return res
        else:
            return [namespace.name for namespace in registry.namespaces]

    async def schema(
        namespace: str, type: str, version: Optional[int]
    ) -> Dict[str, Any]:
        return registry.get(namespace=namespace, type_name=type, version=version)  # type: ignore

    async def schema_registry_enabled() -> bool:
        return True

    async def prewarm_caches() -> None:
        registry.namespaces

except ImportError:

    if configuration.schema_registry_url:
        log.warning(
            (
                "Schema Registry URL configured but schema registry library not available."
                "Registry related features will NOT be available."
            )
        )

    async def namespaces(  # type:ignore
        with_types: bool = False,
    ) -> Union[List[str], Dict]:
        log.debug("Schema Registry not installed. Returning empty result.")
        return []

    async def schema(
        namespace: str, type: str, version: Optional[int]
    ) -> Dict[str, Any]:
        log.debug("Schema Registry not installed. Returning empty result.")
        return {}

    async def schema_registry_enabled() -> bool:
        return False

    async def prewarm_caches() -> None:
        pass



from fastapi import APIRouter, Request
from fastapi.responses import FileResponse

from fact_explorer.app.api.constants import API_RESPONSES
from fact_explorer.config import FRONTEND_SOURCE_PATH
from fact_explorer.config import get_logger


log = get_logger("frontend.router")

router = APIRouter(responses=API_RESPONSES, include_in_schema=False)

# For now as we have only two routes I decided to not do the
# route(/{full_path:path}) magic but rather be explicit in routing
# if we decide to have more pages this decision should be revisited.


@router.get("/")
async def index(request: Request):
    resp = FileResponse(FRONTEND_SOURCE_PATH.joinpath("index.html"))
    return resp


@router.get("/search")
async def search(request: Request):
    resp = FileResponse(FRONTEND_SOURCE_PATH.joinpath("search.html"))
    return resp


@router.get("/favicon.ico")
async def favicon(request: Request):
    resp = FileResponse(FRONTEND_SOURCE_PATH.joinpath("favicon.png"))
    return resp



from typing import Any, Dict
from fastapi import FastAPI


from fact_explorer.app.db.db_session import db_instance as database_instance
from fact_explorer.app.frontend.bridge import SpaStaticFiles
from fact_explorer.app.api.router import router as api_router
from fact_explorer.app.api.registry import router as registry_router
from fact_explorer.app.frontend.router import router as frontend_router
from fact_explorer.app.business.registry import prewarm_caches
from fact_explorer.config import FRONTEND_SOURCE_PATH
from fact_explorer import __version__ as explorer_version
from os import environ as env

app = FastAPI()

# The order of these includes and definitions is important here.
# Later routes will overwrite earlier ones.
app.include_router(api_router)
app.include_router(registry_router)
app.include_router(frontend_router)


@app.get("/health-check")
async def health_check() -> Dict[str, Any]:
    # See: https://tools.ietf.org/id/draft-inadarei-api-health-check-05.html
    return {
        "status": "pass",
        "version": "1",
        "releaseId": str(explorer_version),
        "notes": [""],
        "output": "",
        "checks": {},
    }


app.mount(
    "/",
    SpaStaticFiles(
        directory=str(FRONTEND_SOURCE_PATH),
        html=True,
    ),
    name="Next SPA",
)


@app.on_event("startup")
async def startup() -> None:
    if env.get("DISABLE_CACHE_PREWARM"):
        pass
    else:
        await prewarm_caches()

    await database_instance.connect()
    app.state.db = database_instance



from fact_explorer.config import get_logger, get_configuration

from typing import Any, List, Dict, Optional

log = get_logger("crypto_helpers")
configuration = get_configuration()

try:
    from cryptoshred.asynchronous.convenience import find_and_decrypt
    from cryptoshred.backends import KeyBackend, DynamoDbSsmBackend

    default_key_backend = DynamoDbSsmBackend(
        iv_param=configuration.cryptoshred_init_vector_path
    )

    async def decrypt_result(
        *, data: List[Dict[str, Any]], key_backend: Optional[KeyBackend] = None
    ) -> List[Dict[str, Any]]:
        log.debug("Decrypting data")
        if not key_backend:
            key_backend = default_key_backend
        result = []
        for item in data:
            decrypted = await find_and_decrypt(
                x=item["payload"], key_backend=key_backend
            )

            result.append({"header": item["header"], "payload": decrypted})

        return result

    async def cryptoshred_enabled() -> bool:
        return True

except ImportError:
    KeyBackend = None

    if configuration.cryptoshred_init_vector_path:
        log.warning(
            (
                "Cryptoshred configured but library not available"
                "Cryptoshredding related features will not be available."
            )
        )

    async def decrypt_result(
        *, data: List[Dict[str, Any]], key_backend: Optional[KeyBackend] = None
    ) -> List[Dict[str, Any]]:
        log.debug("Cryptoshred not installed returning raw data")
        return data

    async def cryptoshred_enabled() -> bool:
        return False



import logging
from pathlib import Path

from typing import Tuple, TypeVar, Type, Dict
from os import environ as env

from pydantic import BaseSettings

log = logging.getLogger("fact_explorer")

FRONTEND_SOURCE_PATH = (
    Path(__file__).resolve().parent.joinpath("app/frontend/static").absolute()
)


log_level_from_env = env.get("FACT_EXPLORER_LOG_LEVEL")
if log_level_from_env:

    log_level = getattr(
        logging,
        log_level_from_env.upper(),
    )
    logging.basicConfig(level=log_level)
    log.setLevel(log_level)

C = TypeVar("C", bound="Configuration.Config")


class Configuration(BaseSettings):
    database_password: str
    cryptoshred_init_vector_path: str = ""
    database_user: str = "rdssystem"
    database_host: str = "localhost"
    database_port: str = "5432"
    database_name: str = "postgres"
    schema_registry_url: str = ""

    class Config:
        env_file_encoding = "utf-8"

        @classmethod
        def customize_sources(
            cls: Type[C],
            init_settings: Dict,
            env_settings: Dict,
            file_secret_settings: Dict,
        ) -> Tuple[Dict, Dict, Dict]:
            return (env_settings, init_settings, file_secret_settings)


def get_configuration(
    profile: str = "default",
    *,
    config_dir: Path = Path.home().joinpath(".fact_explorer"),
) -> Configuration:
    log.info("Getting Configuration")

    if profile:
        env_file_location = config_dir.joinpath(f"{profile}.env").absolute()
        return Configuration(_env_file=env_file_location)

    return Configuration()


def get_logger(name: str) -> logging.Logger:
    return log.getChild(name)



from datetime import datetime
from functools import wraps
import asyncpg
from json import dumps, loads
from typing import Any, Callable, Dict, List, Optional

from fact_explorer.config import get_configuration, get_logger


async def _dict_converter(results: Any) -> List[Dict[str, Any]]:
    return [
        {"header": loads(item["header"]), "payload": loads(item["payload"])}
        for item in results
    ]


async def _extract_timestamp(record: Any) -> datetime:
    return datetime.fromtimestamp(int(str(loads(record["header"])["meta"]["_ts"])[:10]))


class Database:
    def __init__(self) -> None:
        self.config = get_configuration()
        self._cursor = None
        self.log = get_logger(self.__class__.__name__)

        self._connection_pool: asyncpg.Pool = None

    def connect_wrapper(func: Callable) -> Callable:  # type: ignore
        @wraps(func)
        async def wrapped(self, *args, **kwargs):  # noqa: ANN
            if not self._connection_pool:
                await self.connect()
            else:
                try:
                    return await func(self, *args, **kwargs)
                except Exception as e:
                    self.log.exception(e)

        return wrapped

    async def connect(self) -> None:
        if not self._connection_pool:
            try:
                self._connection_pool = await asyncpg.create_pool(
                    min_size=1,
                    max_size=10,
                    command_timeout=60,
                    host=self.config.database_host,
                    port=self.config.database_port,
                    user=self.config.database_user,
                    password=self.config.database_password,
                    database=self.config.database_name,
                )

            except Exception as e:
                self.log.exception(e)

    @connect_wrapper
    async def fetch_by_example(
        self,
        *,
        header_example: Optional[Dict[str, str]] = None,
        payload_example: Optional[Dict[str, str]] = None,
        limit: int = 10,
        offset: int = 0
    ) -> List[Dict[str, Any]]:
        if not header_example:
            header_example = {}
        if not payload_example:
            payload_example = {}

        con = await self._connection_pool.acquire()

        try:
            stmt = await con.prepare(  # type: ignore
                (
                    "select *"
                    " from fact"
                    " where header @> cast($1 as jsonb)"
                    " and payload @> cast ($2 as jsonb)"
                    " order by ser desc"
                    " limit $3::int"
                    " offset $4::int"
                )
            )
            self.log.debug(stmt.get_query())
            result = await stmt.fetch(
                dumps(header_example), dumps(payload_example), limit, offset
            )
            return await _dict_converter(result)
        finally:
            await self._connection_pool.release(con)

    @connect_wrapper
    async def fetch_by_time(self, *, until: datetime) -> List[Dict[str, Any]]:
        """
        Will fetch the entries up to approximately until. Will fetch at least 50 rows
        iff available. Will fetch in batches afterwards. Since factcast does not have
        an index on _ts this is faster than a FTS although of course less than ideal
        """
        result = []
        con = await self._connection_pool.acquire()

        try:
            async with con.transaction():  # type: ignore
                cursor = await con.cursor("select *" " from fact" " order by ser desc")  # type: ignore
                result += await cursor.fetch(50)
                last_retrieved_time = await _extract_timestamp(record=result[-1])
                while last_retrieved_time > until:
                    # Doing this by fetching rows. it is way faster than a FTS.
                    result += await cursor.fetch(50)
                    last_retrieved_time = await _extract_timestamp(record=result[-1])
                return await _dict_converter(result)
        finally:
            await self._connection_pool.release(con)


db_instance = Database()
